import typing

T = typing.TypeVar("T")


class classproperty:
    def __init__(self, fget: typing.Callable[[typing.Type[T]], typing.Any]) -> None:
        """
        Initialize the classproperty.

        Args:
            fget: A function that takes the class as an argument and returns a value.
        """
        self.fget = fget

    def __get__(self, obj: typing.Any, owner: typing.Type[T]) -> typing.Any:
        """
        Retrieve the property value.

        Args:
            obj: The instance of the class (unused).
            owner: The class that owns the property.

        Returns:
            The value returned by the function.
        """
        return self.fget(owner)


# something like @abstractclassproperty doesn't work, otherwise pycharm won't detect it as abstract!



import os

import pytest
from typedal import TypeDAL

from src.edwh_migrate import activate_migrations, setup_db

from .fixtures import (  # noqa
    clean_migrate,
    fixture_temp_chdir,
    sqlite_empty,
    tmp_empty_sqlite_db_file,
    tmp_just_implemented_features_sqlite_db_file,
    tmp_just_implemented_features_sqlite_sql_file,
    tmp_sqlite_folder,
    tmp_sqlite_sql_file,
)


@pytest.fixture
def env_use_typedal():
    os.environ["USE_TYPEDAL"] = "1"
    yield
    del os.environ["USE_TYPEDAL"]


@pytest.fixture
def tmp_typedal_env(fixture_temp_chdir):
    os.environ["DB_URI"] = f"sqlite://{fixture_temp_chdir / 'some.sqlite'}"
    os.environ["FOLDER"] = str(fixture_temp_chdir / "database")
    yield
    del os.environ["DB_URI"]
    del os.environ["FOLDER"]


def test_setup_db(tmp_just_implemented_features_sqlite_db_file, clean_migrate):
    db = setup_db(dal_class=TypeDAL)

    assert isinstance(db, TypeDAL)


def test_setup_empty(tmp_typedal_env, clean_migrate):
    db = setup_db(dal_class=TypeDAL, migrate=True, migrate_enabled=True)

    assert isinstance(db, TypeDAL)


def test_activate_migrations(env_use_typedal, tmp_typedal_env, clean_migrate):
    # no migrations -> return false
    assert activate_migrations() is False



import os
from sqlite3 import OperationalError

import pytest

from src.edwh_migrate import (
    ViewMigrationManager,
    migrate,
)

from .fixtures import clean_migrate, sqlite_empty, tmp_empty_sqlite_db_file, tmp_sqlite_folder  # noqa


class StandaloneView(ViewMigrationManager):
    since = "first_migration"
    until = "third_migration"  # doesn't exist

    def up(self): ...

    def down(self): ...


class Ignored(ViewMigrationManager):
    since = "second_migration"  # doesn't exist

    def up(self):
        2 / 0

    def down(self):
        1 / 0


class MyBaseView(ViewMigrationManager):
    uses = ()

    def up(self) -> None:
        self.db.executesql("""
            CREATE VIEW base_view AS
            SELECT Rowid AS id
            ;
        """)

    def down(self) -> None:
        self.db.executesql("""
            DROP VIEW IF EXISTS base_view;
        """)


class MyChildView1(ViewMigrationManager):
    uses = (MyBaseView,)

    def up(self) -> None:
        self.db.executesql("""
            CREATE VIEW my_child_view_1 AS
            SELECT id FROM base_view
            ;
        """)

    def down(self) -> None:
        self.db.executesql("""
            DROP VIEW IF EXISTS my_child_view_1;
        """)


class MyChildView2(ViewMigrationManager):
    uses = (MyBaseView, MyChildView1)

    def up(self) -> None:
        self.db.executesql("""
            CREATE VIEW my_child_view_2 AS
            SELECT id FROM my_child_view_1
            ;
        """)

    def down(self) -> None:
        self.db.executesql("""
            DROP VIEW IF EXISTS my_child_view_2;
        """)


def test_resolving_manager_order(tmp_empty_sqlite_db_file, clean_migrate):
    @migrate.migration()
    def first_migration(db):
        with StandaloneView(db), Ignored(db):
            ...

        return True

    config = migrate.get_config()

    assert len(migrate.migrations) == 1
    assert migrate.activate_migrations(config=config)

    db = migrate.setup_db(config=config)

    assert db(db.ewh_implemented_features).count()

    # one should be fine:
    with MyChildView2(db):
        ...

    with MyBaseView(db):
        ...

    # multiple without combine may crash:
    with pytest.raises(OperationalError):
        with MyChildView2(db), MyBaseView(db), MyChildView1(db):
            ...



import shutil
import tempfile
from pathlib import Path

import pytest
from configuraptor import Singleton
from pydal import DAL
from testcontainers.postgres import PostgresContainer

from src.edwh_migrate import (
    Config,
    activate_migrations,
    migrate,
    recover_database_from_backup,
)
from src.edwh_migrate.postgres import PostgresUndefinedTable


def rmdir(path: Path):
    shutil.rmtree(path)


DB_NAME = "edwh_migrate_test"

postgres = PostgresContainer("postgres:16-alpine", dbname=DB_NAME)


@pytest.fixture(scope="module", autouse=True)
def psql(request):
    # defer teardown:
    request.addfinalizer(postgres.stop)

    postgres.start()
    # note: ONE PostgresContainer with scope module can be used,
    # if you try to use containers in a function scope, it will not work.
    # thus, this clean_db fixture is added to cleanup between tests:


@pytest.fixture(scope="function", autouse=True)
def clean_db():
    # Connection URI
    from sqlalchemy_utils.functions import create_database, drop_database

    uri = postgres.get_connection_url()
    drop_database(uri)
    create_database(uri)

    Singleton.clear()


@pytest.fixture()
def conn_str():
    conn_str = postgres.get_connection_url()
    # make pydal-friendly:
    return "postgres://" + conn_str.split("://")[-1]


@pytest.fixture()
def tempdir():
    with tempfile.TemporaryDirectory() as d:
        yield d


def test_setup_on_psql_not_long_running(conn_str: str, tempdir: str):
    config = Config.load(dict(migrate_uri=conn_str, db_folder=tempdir))

    assert config.migrate_uri.startswith("postgres://")

    db = migrate.setup_db(
        config=config, migrate=True, migrate_enabled=True, long_running=False, remove_migrate_tablefile=True
    )

    assert db.ewh_implemented_features


def test_setup_on_psql_long_running(conn_str: str, tempdir: str):
    config = Config.load(dict(migrate_uri=conn_str, db_folder=tempdir))

    db = migrate.setup_db(
        config=config, migrate=True, migrate_enabled=True, long_running=True, remove_migrate_tablefile=True
    )

    assert db.ewh_implemented_features


def psql_backup():
    path = Path(__file__).parent / "sqlite_empty" / "just_implemented_features.psql.sql"
    return str(path)


def test_postgres_backup(conn_str: str, tempdir: str):
    db = DAL(conn_str)

    with pytest.raises(PostgresUndefinedTable):
        db.executesql("SELECT count(*) FROM ewh_implemented_features")
    db.rollback()

    config = Config.load(dict(migrate_uri=conn_str, db_folder=tempdir, database_to_restore=psql_backup()))
    recover_database_from_backup(set_schema="public")

    rows = db.executesql("SELECT count(*) FROM ewh_implemented_features")
    assert rows[0][0] == 0  # no rows, but table exists


def test_postgres_unavailable(conn_str: str, tempdir: str):
    config = Config.load(
        dict(
            migrate_uri=conn_str.replace(DB_NAME, "INVALID_DB_NAME"),
            db_folder=tempdir,
            database_to_restore=psql_backup(),
        )
    )
    with pytest.raises(ValueError):
        activate_migrations(config=config, max_time=5)



# coding=utf-8
# SPDX-FileCopyrightText: 2023-present Remco <remco@educationwarehouse.nl>
#
# SPDX-License-Identifier: MIT
# type: ignore
import os

from edwh_migrate import ViewMigrationManager, migration

if os.environ.get("PYTEST_CURRENT_TEST"):
    raise EnvironmentError("This file is not supposed to be executed by pytest!")


class ExampleDependency(ViewMigrationManager):
    # will also run used_by (= ExampleViewManager)
    uses = ()
    since = "feature_3"

    def down(self):
        print("2. this happens before before the migration", self.db._uri)

    def up(self):
        print("3. this happens after after the migration", self.db._uri)


class ExampleViewManager(ViewMigrationManager):
    uses = [
        ExampleDependency,
    ]
    # only runs itself, not the dependency

    def down(self):
        print("1. this happens before the migration", self.db._uri)

    def up(self):
        print("4. this happens after the migration", self.db._uri)


@migration
def feature_1(db):
    # mogelijkheden om in een feature te doen:
    # 1. SQL strings uitvoeren richting database
    # 2. Bestanden lezen en uitvoeren als sql string, of verschillende statements achter elkaar
    # 3. PG_RESTORE uitvoeren mbv plumbum om zo een hele database te recoveren
    # en daarnaast
    # 1. wijzigingen die in de data horen nav een schemawijziging...
    #
    # een andere manier zou zijn om dit uit bestanden te lezen en die door te voeren
    # dan heb je de functies niet zo zeer nodig, maar heb je wel je schema bijgewerkt. Zolang je geen
    # logica van python nodig hebt kun je best veel doen met alleen maar SQL scripts
    # sorted(/features/*.sql) | cat vanaf invoer |
    return True


@migration(requires=feature_1)
def feature_2(db):
    print("feature_2")
    return True


@migration(requires=feature_2)
def feature_3(db):
    print("feature_3")
    return True


@migration
def functionalname_date_sequencenr(db):
    with ExampleDependency(db):
        db.executesql("""
        
        """)

    db.commit()
    return False
